#pragma once

/*
 * Odyssey.
 *
 * Scalable PostgreSQL connection pooler.
 * 
 * general definitions for all odyssey source files
 * should be included first
 */

#include <od_c.h>

#include <build.h>
